/*-----------------------------------------------------------------------------

 2002, Jeff M Soft und Hardware GmbH, All Rights Reserved

-----------------------------------------------------------------------------*/

#include "Module.h"
#include "SEMod_struct.h"
#include "SEPin.h"
#include <assert.h>
//#include <iostream>
//#include <fstream>

// define some constants to make referencing in/outs clearer
#define PN_OUTPUT1 0
#define PN_INPUT1 1

// The most important part, processing the audio
void Module::sub_process(long buffer_offset, long sampleFrames )
{
	// assign some pointers to your in/output buffers. usually blocks (array) of 96 samples
    float *out1 = buffer_offset + output1;

	/*
	// first - simple - version, not optimized
	for(int s = 0 ; s < sampleFrames ; s++ ) 
	{
		// step though each input, calculating the average
		float product = 1.f;
		for(int i = 0 ; i < dynamic_plugs_count; i ++ ) 
		{ 
			float *inp = s + buffer_offset + dynamic_plugs[i];
			product = product * *inp * 10.f;
		}

		*out1++ = product * 0.1f;
	} 
	*/
	
	// second version, faster
	switch (dynamic_plugs_count)
	{
	case 1:
		{
			float *inp = buffer_offset + dynamic_plugs[0];
			for(int s = 0 ; s < sampleFrames ; s++ ) *out1++ = *inp++;
		}
		break;

	case 2:
		{
			float *inp1 = buffer_offset + dynamic_plugs[0];
			float *inp2 = buffer_offset + dynamic_plugs[1];
			for(int s = 0 ; s < sampleFrames ; s++ ) 
				*out1++ = *inp1++ * *inp2++ * 10.f;
		}
		break;
	
	case 3:
		{
			float *inp1 = buffer_offset + dynamic_plugs[0];
			float *inp2 = buffer_offset + dynamic_plugs[1];
			float *inp3 = buffer_offset + dynamic_plugs[2];
			for(int s = 0 ; s < sampleFrames ; s++ ) 
				*out1++ = *inp1++ * *inp2++ * *inp3++ * 100.f;
		}
		break;
	
	case 4:
		{
			float *inp1 = buffer_offset + dynamic_plugs[0];
			float *inp2 = buffer_offset + dynamic_plugs[1];
			float *inp3 = buffer_offset + dynamic_plugs[2];
			float *inp4 = buffer_offset + dynamic_plugs[3];
			for(int s = 0 ; s < sampleFrames ; s++ ) 
				*out1++ = *inp1++ * *inp2++ * *inp3++ * *inp4++ * 1000.f;
		}
		break;

	case 5:
		{
			float *inp1 = buffer_offset + dynamic_plugs[0];
			float *inp2 = buffer_offset + dynamic_plugs[1];
			float *inp3 = buffer_offset + dynamic_plugs[2];
			float *inp4 = buffer_offset + dynamic_plugs[3];
			float *inp5 = buffer_offset + dynamic_plugs[4];
			for(int s = 0 ; s < sampleFrames ; s++ ) 
				*out1++ = *inp1++ * *inp2++ * *inp3++ * *inp4++ * *inp5++ * 10000.f;
		}
		break;

	case 6:
		{
			float *inp1 = buffer_offset + dynamic_plugs[0];
			float *inp2 = buffer_offset + dynamic_plugs[1];
			float *inp3 = buffer_offset + dynamic_plugs[2];
			float *inp4 = buffer_offset + dynamic_plugs[3];
			float *inp5 = buffer_offset + dynamic_plugs[4];
			float *inp6 = buffer_offset + dynamic_plugs[5];
			for(int s = 0 ; s < sampleFrames ; s++ ) 
				*out1++ = *inp1++ * *inp2++ * *inp3++ * *inp4++ * *inp5++ * *inp6++ * 100000.f;
		}
		break;

	case 7:
		{
			float *inp1 = buffer_offset + dynamic_plugs[0];
			float *inp2 = buffer_offset + dynamic_plugs[1];
			float *inp3 = buffer_offset + dynamic_plugs[2];
			float *inp4 = buffer_offset + dynamic_plugs[3];
			float *inp5 = buffer_offset + dynamic_plugs[4];
			float *inp6 = buffer_offset + dynamic_plugs[5];
			float *inp7 = buffer_offset + dynamic_plugs[6];
			for(int s = 0 ; s < sampleFrames ; s++ ) 
				*out1++ = *inp1++ * *inp2++ * *inp3++ * *inp4++ * *inp5++ * *inp6++ 
					* *inp7++ * 1000000.f;
		}
		break;

	case 8:
		{
			float *inp1 = buffer_offset + dynamic_plugs[0];
			float *inp2 = buffer_offset + dynamic_plugs[1];
			float *inp3 = buffer_offset + dynamic_plugs[2];
			float *inp4 = buffer_offset + dynamic_plugs[3];
			float *inp5 = buffer_offset + dynamic_plugs[4];
			float *inp6 = buffer_offset + dynamic_plugs[5];
			float *inp7 = buffer_offset + dynamic_plugs[6];
			float *inp8 = buffer_offset + dynamic_plugs[7];
			for(int s = 0 ; s < sampleFrames ; s++ ) 
				*out1++ = *inp1++ * *inp2++ * *inp3++ * *inp4++ * *inp5++ * *inp6++ 
					* *inp7++ * *inp8++ * 10000000.f;
		}
		break;

	default:
		for(int	s = 0 ; s < sampleFrames ; s++ ) 
		{
			// step though each input, calculating the average
			float product = 1.f;
			for(int i = 0 ; i < dynamic_plugs_count; i ++ ) 
			{ 
				float *inp = s + buffer_offset + dynamic_plugs[i];		
				product = product * *inp * 10.f;
			}
			*out1++ = product * 0.1f;
		}
	} 
	

	/*
	// another version, a bit faster than the first
	for(int s = 0 ; s < sampleFrames ; s++ ) 
	{
		// step though each input, calculating the average
		float product = 1.f;
		for(int i = 0 ; i < dynamic_plugs_count; i ++ ) 
		{ 
			float *inp = s + buffer_offset + dynamic_plugs[i];
			product = product * *inp;
		}

		static const float coeff_table[] = 
		{
			1.f,
			10.f,
			100.f,
			1000.f,
			10000.f,
			100000.f,
			1000000.f,
			10000000.f,
			100000000.f,
			1000000000.f,
			10000000000.f,
			100000000000.f,
			1000000000000.f
		};

		*out1++ = product * coeff_table[dynamic_plugs_count];
	*/

}

void Module::sub_process_static(long buffer_offset, long sampleFrames )
{
	sub_process(buffer_offset, sampleFrames);
	static_count = static_count - sampleFrames;
	if( static_count <= 0 )
	{
		CallHost(seaudioMasterSleepMode);
	}
}

Module::Module(seaudioMasterCallback2 seaudioMaster, void *p_resvd1) : SEModule_base(seaudioMaster, p_resvd1)
,dynamic_plugs(0)
{
}

void Module::open()
{
	// choose which function is used to process audio
	SET_PROCESS_FUNC(Module::sub_process);

	// call the base class
	SEModule_base::open();
	// to work out how many 'dynamic' plugs the module has..
	// Ask host how many input plugs this module actually has,
	// then subtract the number of 'regular' plugs
	dynamic_plugs_count = CallHost(seaudioMasterGetTotalPinCount) - PN_INPUT1;

	if( dynamic_plugs_count > 0 )
	{
		// allocate an array to hold pointers to all the input buffers
		dynamic_plugs = new float*[dynamic_plugs_count]; 

		// ask the host for a pointer to each input buffer
		// store them in the array
		for(int i = 0 ; i < dynamic_plugs_count ; i++ ) 
		{ 
			dynamic_plugs[i] = *(float **) getPin(i + PN_INPUT1)->GetVariableAddress(); //(float*)CallHost(seaudioMasterGetPinVarAddress, i + PN_INPUT1); 
		}
	}
}

Module::~Module()
{
	// This is where you free any memory/resources your module has created
	delete [] dynamic_plugs;
}

// describe your module
bool Module::getModuleProperties (SEModuleProperties* properties)
{
	// describe the plugin, this is the name the end-user will see.
	properties->name = "sc:CascMult";								// !!TODO!!

	// return a unique string 32 characters max
	// if posible include manufacturer and plugin identity
	// this is used internally by SE to identify the plug.
	// No two plugs may have the same id.
	properties->id = "sc:CascMult";						// !!TODO!!

	// Info, may include Author, Web page whatever
	properties->about = "Scoofster Cascaded Multiplier v1.2";								// !!TODO!!
	return true;
}

// describe the pins (plugs and parameters)
bool Module::getPinProperties (long index, SEPinProperties* properties)
{
	switch( index )								// !!TODO!! list your in / out plugs
	{
	case 0:
		properties->name				= "Output";
		properties->variable_address	= &output1;
		properties->direction			= DR_OUT;
		properties->datatype			= DT_FSAMPLE;
		break;
	case 1: // this plug automatically duplicates itself
			// it must be the last plug in the list
		properties->name				= "Input";
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		properties->flags = IO_AUTODUPLICATE | IO_CUSTOMISABLE;
		break;
	default:
		return false; // host will ask for plugs 0,1,2,3 etc. return false to signal when done
	};

	return true;
}

// An input plug has changed value
void Module::OnPlugStateChange(SEPin *pin)
{
	// query the 'state of the input plugs...
	//     ST_RUN     = Normal Streaming Audio        (e.g. from an oscillator)
	//     ST_STATIC  = Fixed, unchanging input value (e.g. a slider at rest)

	// we need to pass on the state of this module's output signal
	// it depends on the inputs.  Choose the 'highest'..

	state_type out_stat = ST_STATIC;
	
		//std::ofstream tx;
		//tx.open("f:/debug.txt", std::ios::app );
	
		//tx << "start OnPlugStateChange" << std::endl;
	
	for(int i=0; i<dynamic_plugs_count; i++ ) 
	{		
		SEPin *p = getPin(PN_INPUT1 + i);
		state_type in_stat = p->getStatus();

		/*
			tx << SampleClock() << ", " << i <<  ", ";
			if (in_stat == ST_STATIC) tx << "ST_STATIC";
			else if (in_stat == ST_RUN) tx << "ST_RUN";
			else tx << "status=" << in_stat;
			tx << std::endl;
		*/
			
		/*
			tx << "timestamp:" << SampleClock() << std::endl;
			tx << "i = " << i << std::endl;
			tx << "status = ";
			tx << "pin no.: " << PN_INPUT1 + i << std::endl;
			if (in_stat == ST_STATIC) tx << "ST_STATIC";
			else tx << "ST_RUN";
			tx << std::endl;
			tx << std::endl;
		*/

		if (in_stat <= ST_STATIC && p->getValue()==0.f)
		{
			out_stat = ST_STATIC;
			break; break;
		}
		else if( in_stat > out_stat ) 
		{
			out_stat = in_stat;
		}

	}
	
		//tx.close();

	// 'transmit' this modules new output status to next module 'downstream'
	getPin(PN_OUTPUT1)->TransmitStatusChange( SampleClock(), out_stat);

	// setup 'sleep mode' or not
	if( out_stat < ST_RUN ) 
	{
		static_count = getBlockSize();
		SET_PROCESS_FUNC(Module::sub_process_static);
	}
	else
	{
		SET_PROCESS_FUNC(Module::sub_process);
	}
/*
	// iterate through all clones example code
	SEMod_struct_base *clone_struct;
	// get first one
	CallHost(seaudioMasterGetFirstClone,0,0,&clone_struct);

	while( clone_struct != 0 )
	{
        Module *clone = ((Module *)(clone_struct->object));

		// Access each clone here

		// step to next clone
		clone->CallHost(seaudioMasterGetNextClone,0,0,&clone_struct);
	}
*/
}
